import shutil
from pathlib import Path
from textwrap import dedent

import mammos_dft
import mammos_entity as me
import pytest
import yaml

from mammos_spindynamics import uppasd
from mammos_spindynamics.uppasd import _inpsd, _simulation


def test_serialise_cell():
    assert (
        _inpsd._serialise_cell([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        == "1 2 3\n      4 5 6\n      7 8 9"
    )
    assert (
        _inpsd._serialise_cell([[1.0, 2, 3], [4, 5, 6], [7, 8, 9]])
        == "1.0 2.0 3.0\n      4.0 5.0 6.0\n      7.0 8.0 9.0"
    )

    with pytest.raises(ValueError):
        _inpsd._serialise_cell([1, 2, 3])

    with pytest.raises(TypeError):
        _inpsd._serialise_cell("abc")


def test_serialize_ncell():
    assert _inpsd._serialize_array("ncell", [12, 20, 1]) == "12 20 1"
    with pytest.raises(ValueError):
        _inpsd._serialize_array("ncell", [12, 20])


@pytest.mark.skip(reason="not yet implemented")
def test_create_input_files():
    pass


@pytest.mark.skip(reason="not yet implemented")
def test_preprocess_inpsd_dat():
    pass


def test_create_run_dir(tmp_path: Path):
    """Create a new directory for a run.

    If runs exist in the directory the index is increased
    """
    path, index = _simulation._create_run_dir(tmp_path)
    assert path == tmp_path / "0-run"
    assert index == 0

    path, index = _simulation._create_run_dir(tmp_path)
    assert path == tmp_path / "1-run"
    assert index == 1

    path, index = _simulation._create_run_dir(tmp_path, mode="temperature_sweep")
    assert path == tmp_path / "2-temperature_sweep"
    assert index == 2

    (tmp_path / "3-some-other-file").touch()

    path, index = _simulation._create_run_dir(tmp_path)
    assert path == tmp_path / "3-run"
    assert index == 3

    with pytest.raises(ValueError):
        _simulation._create_run_dir(tmp_path, mode="unsupported")

    path, index = _simulation._create_run_dir(tmp_path / "out")
    assert path == tmp_path / "out" / "0-run"
    assert index == 0

    (tmp_path / "out" / "1-run").touch()

    path, index = _simulation._create_run_dir(tmp_path / "out")
    assert path == tmp_path / "out" / "2-run"
    assert index == 2

    (tmp_path / "a_file").touch()
    with pytest.raises(RuntimeError):
        path, index = _simulation._create_run_dir(tmp_path / "a_file")

    path, index = _simulation._create_run_dir(
        tmp_path / "out2", mode="temperature_sweep"
    )
    assert path == tmp_path / "out2" / "0-temperature_sweep"
    assert index == 0


def test_write_inputs(tmp_path: Path):
    """Write all inputs to existing run directory.

    - a new file inpsd.dat is created with content inp_file_content
    - all files in 'files_to_copy' are copied to the run dir, the new file name is the
      dict key
    - a new file mammos_spindynamics.yaml is created with content metadata
    """
    inp_file_content = "the\ninp file"
    files_to_copy = {
        "exchange": tmp_path / "exchange",
        "momfile": tmp_path / "common" / "momfile",
    }
    metadata = {"fake_data": True}

    run_path = tmp_path / "out" / "run-0"

    # output directory does not exist; created in a prior function
    with pytest.raises(FileNotFoundError, match="inpsd.dat"):
        _simulation._write_inputs(run_path, inp_file_content, files_to_copy, metadata)

    run_path.mkdir(parents=True)

    # input files missing
    with pytest.raises(FileNotFoundError, match="exchange"):
        _simulation._write_inputs(run_path, inp_file_content, files_to_copy, metadata)

    for path in files_to_copy.values():
        path.parent.mkdir(exist_ok=True)
        path.touch()

    _simulation._write_inputs(run_path, inp_file_content, files_to_copy, metadata)
    assert (run_path / "inpsd.dat").read_text() == "the\ninp file"
    assert (run_path / "exchange").exists()
    assert (run_path / "momfile").exists()
    assert (run_path / "mammos_spindynamics.yaml").exists()
    with open(run_path / "mammos_spindynamics.yaml") as f:
        assert yaml.safe_load(f) == metadata


def test_find_executable():
    exe = _simulation._find_executable("uppasd")
    assert isinstance(exe, Path)
    assert ".pixi/envs/default/" in exe.as_posix()  # common part on Windows and Unix
    with pytest.raises(RuntimeError):
        _simulation._find_executable("this-is-not-uppasd")


@pytest.mark.skip(reason="not yet implemented")
def test_update_metadata_file():
    pass


# === Integration tests using public interface ===


def test_simulation_run(tmp_path: Path):
    uppasd_inputs = mammos_dft.db.get_uppasd_properties("Co2Fe2H4")
    sim = uppasd.Simulation()

    with pytest.raises(RuntimeError, match="parameters are missing"):
        sim.run(out=tmp_path)

    sim = uppasd.Simulation(
        cell=[(1.0, -0.5, 0), (0, 0.866, 0), (0, 0, 3.228)],
        ncell=[12, 12, 12],
        posfiletype="D",
        maptype=2,
        initmag=4,
        alat=2.65e-10,
        ip_mcnstep=100,
        mcnstep=100,
        exchange=uppasd_inputs.exchange,
        momfile=uppasd_inputs.momfile,
    )

    with pytest.raises(RuntimeError, match="parameters are missing"):
        sim.run(out=tmp_path)

    with pytest.raises(AttributeError, match=r"Missing argument 'posfile'."):
        sim.run(out=tmp_path, ip_temp=50, temp=70)

    with pytest.raises(AttributeError, match="restartfile for initmag 4 missing"):
        sim.run(
            out=tmp_path,
            ip_temp=50,
            temp=70,
            posfile=uppasd_inputs.posfile,
        )

    res = sim.run(
        out=tmp_path,
        ip_temp=50,
        temp=70,
        posfile=uppasd_inputs.posfile,
        initmag=3,
    )
    assert isinstance(res, uppasd.RunData)
    assert me.T(70) == res.T
    assert res.Ms.value > 0
    assert (res.out / "uppasd_stdout.txt").is_file()
    assert (res.out / "uppasd_stderr.txt").is_file()


@pytest.mark.timeout(30)
def test_simulation_run_with_modified_inpsd_dat(tmp_path: Path):
    """Start simulation from an inpsd.dat file.

    The file sets a large number of MC steps, which are overwritten in run. If the
    overwrite would not work, we would run into the 30 second timeout.
    """
    # input file for Co2Fe2H4 taken from https://github.com/mammos-project/uppsala-data
    # - number of cells reduced to speed up simulation
    # - initmag placeholder replaced with 3
    (tmp_path / "input").write_text(
        dedent(
            """\
            simid UppASD__123

            cell  1.000000000000000  -0.500000000182990   0.000000000000000
                0.000000000000000   0.866011964524435   0.000000000000000
                0.000000000000000   0.000000000000000   3.228114436804486

            ncell 12  12  12
            bc    P P P
            sym 0
            posfile ./posfile
            posfiletype D

            initmag 3
            momfile ./momfile
            maptype 2
            exchange ./exchange
            ip_mode M
            ip_temp TEMP
            ip_mcnstep 25000

            mode M
            temp TEMP
            mcnstep 50000
            plotenergy 1
            do_proj_avrg Y
            do_cumu Y

            alat 2.6489381562e-10"""
        )
    )

    uppasd_inputs = mammos_dft.db.get_uppasd_properties("Co2Fe2H4")
    sim = uppasd.Simulation(
        inpsd=tmp_path / "input",
        exchange=uppasd_inputs.exchange,
        momfile=uppasd_inputs.momfile,
        posfile=uppasd_inputs.posfile,
    )

    # temperature is still set to the placeholder; uppasd prints warnings but still
    # runs a simulation and returns 0; we only warn that we have found ERROR in stdout;
    # the full stdout is shown to the user in the warning message
    with pytest.warns(UserWarning, match="UppASD output contains ERROR lines"):
        sim.run(tmp_path / "my" / "output", ip_mcnstep=100, mcnstep=100)

    res = sim.run(tmp_path / "my" / "output", ip_mcnstep=100, mcnstep=100, T=10)
    assert isinstance(res, uppasd.RunData)
    assert me.T(10) == res.T
    assert res.Ms.value > 0

    with pytest.warns(UserWarning, match="parameter 'restartfile' is ignored"):
        sim.run(
            out=tmp_path / "my" / "output",
            ip_mcnstep=50,
            mcnstep=50,
            T=10,
            initmag=3,
            restartfile=res.restartfile,
        )


def test_simulation_run_with_unmodified_inpsd_dat(tmp_path: Path):
    """Run simulation for an inpsd.dat file without any modifications.

    Without *any* modifications is actually not true, the code will always rewrite
    the lines for posfile, momfile, and exchange to guarantee that the run directory
    is fully self-contained.
    """
    (tmp_path / "inpsd.dat").write_text(
        dedent(
            """\
            simid UppASD__123

            cell  1.000000000000000  -0.500000000182990   0.000000000000000
                0.000000000000000   0.866011964524435   0.000000000000000
                0.000000000000000   0.000000000000000   3.228114436804486

            ncell 12  12  12
            bc    P P P
            sym 0
            posfile ./posfile
            posfiletype D

            initmag 3
            momfile ./momfile
            maptype 2
            exchange ./exchange
            ip_mode M
            ip_temp 20
            ip_mcnstep 120

            mode M
            temp 20
            mcnstep 120
            plotenergy 1
            do_proj_avrg Y
            do_cumu Y

            alat 2.6489381562e-10"""
        )
    )
    uppasd_inputs = mammos_dft.db.get_uppasd_properties("Co2Fe2H4")

    for file in ["exchange", "momfile", "posfile"]:
        shutil.copy(getattr(uppasd_inputs, file), file)
    sim = uppasd.Simulation(
        inpsd=tmp_path / "inpsd.dat",
    )
    res = sim.run(tmp_path / "simulation-output")
    assert isinstance(res, uppasd.RunData)
    assert me.T(20) == res.T
    assert res.Ms.value > 0



"""Module for interfacing with UppASD."""

from mammos_spindynamics.uppasd._data import (
    MammosUppasdData,
    RunData,
    TemperatureSweepData,
    read,
)

from ._simulation import Simulation

__all__ = [
    "MammosUppasdData",
    "RunData",
    "Simulation",
    "TemperatureSweepData",
    "read",
]



"""Test uppasd interface."""

import pathlib

import mammos_entity as me
import numpy as np
import pandas as pd
import yaml

from mammos_spindynamics import uppasd


def test_read_function(DATA):
    """Test `read` function.

    The `mammos_spindynamics.uppasd.read` function should understand if the
    given input defines a `MammosUppasdData`, a `RunData`, or a `TemperatureSweepData`.
    """
    mammos_uppasd_data = uppasd.read(DATA / "uppasd")
    assert isinstance(mammos_uppasd_data, uppasd.MammosUppasdData)
    run_data = uppasd.read(DATA / "uppasd" / "1-run")
    assert isinstance(run_data, uppasd.RunData)
    temperature_sweep_data = uppasd.read(DATA / "uppasd" / "0-temperature_sweep")
    assert isinstance(temperature_sweep_data, uppasd.TemperatureSweepData)
    sub_run_data = uppasd.read(DATA / "uppasd" / "0-temperature_sweep" / "0-run")
    assert isinstance(sub_run_data, uppasd.RunData)


def test_MammosUppasdData_class(DATA):
    """Test MammosUppasdData class."""
    mammos_uppasd_data = uppasd.read(DATA / "uppasd")
    assert mammos_uppasd_data.out == pathlib.Path(DATA / "uppasd")
    info_df = pd.DataFrame(
        {
            "name": ["0-temperature_sweep", "1-run"],
            "description": ["Test sweep.", "Single run."],
            "time_elapsed": [None, "1 day, 1:01:01"],
            "T": [[2, 5], 10],
        },
        index=[0, 1],
    )
    assert info_df.equals(mammos_uppasd_data.info())
    assert isinstance(mammos_uppasd_data[0], uppasd.TemperatureSweepData)
    assert isinstance(mammos_uppasd_data[1], uppasd.RunData)


def test_MammosUppasdData_yaml(DATA):
    """Test MammosUppasdData yaml file."""
    with open(DATA / "uppasd" / "mammos_spindynamics.yaml") as f:
        info = yaml.safe_load(f)
    assert info == {
        "metadata": {"mode": "mammos_uppasd_data"},
    }


def test_RunData_class(DATA):
    """Test RunData class."""
    run_dir = DATA / "uppasd" / "1-run"
    run_data = uppasd.read(run_dir)
    assert run_data.out == pathlib.Path(run_dir)
    assert run_data.metadata == {
        "index": 1,
        "description": "Single run.",
        "UppASD version": "v6.0.2",
        "mammos_spindynamics_version": "0.2.5",
        "mode": "run",
        "time_elapsed": "1 day, 1:01:01",
        "time_end": "2025-03-15T13:04:26+01:00",
        "time_start": "2025-03-14T12:03:25+01:00",
    }
    assert run_data.parameters == {"T": 10}

    info_df = pd.DataFrame(
        {
            "name": ["1-run"],
            "description": ["Single run."],
            "time_elapsed": ["1 day, 1:01:01"],
            "T": [10],
        }
    )
    assert info_df.equals(run_data.info())
    assert me.T(10) == run_data.T
    assert me.Ms(6834.473593675746) == run_data.Ms
    assert me.Entity("IsochoricHeatCapacity", 0) == run_data.Cv
    assert me.Entity("Energy", 0) == run_data.E
    assert run_data.U_binder == 0.512274383
    assert run_data.inpsd == pathlib.Path(run_dir / "inpsd.dat")
    assert run_data.exchange == pathlib.Path(run_dir / "jfile")
    assert run_data.momfile == pathlib.Path(run_dir / "momfile")
    assert run_data.posfile == pathlib.Path(run_dir / "posfile")
    assert run_data.cumulants == pathlib.Path(run_dir / "cumulants.bccFe100.out")
    assert run_data.restartfile == pathlib.Path(run_dir / "restart.bccFe100.out")
    assert run_data.n_magnetic_atoms == 9


def test_RunData_yaml(DATA):
    """Test RunData yaml file."""
    with open(DATA / "uppasd" / "1-run" / "mammos_spindynamics.yaml") as f:
        info = yaml.safe_load(f)
    assert info == {
        "metadata": {
            "index": 1,
            "description": "Single run.",
            "UppASD version": "v6.0.2",
            "mammos_spindynamics_version": "0.2.5",
            "mode": "run",
            "time_elapsed": "1 day, 1:01:01",
            "time_end": "2025-03-15T13:04:26+01:00",
            "time_start": "2025-03-14T12:03:25+01:00",
        },
        "parameters": {"T": 10},
    }


def test_TemperatureSweepData_class(DATA):
    """Test TemperatureSweepData class."""
    sweep_dir = DATA / "uppasd" / "0-temperature_sweep"
    sweep_data = uppasd.read(sweep_dir)
    assert sweep_data.out == pathlib.Path(sweep_dir)
    assert sweep_data.metadata == {
        "index": 0,
        "description": "Test sweep.",
        "UppASD version": "v6.0.2",
        "mammos_spindynamics_version": "0.2.5",
        "mode": "temperature_sweep",
    }
    assert sweep_data.parameters == {"T": [2, 5]}

    info_df = pd.DataFrame(
        {
            "name": ["0-run", "1-run"],
            "description": ["", ""],
            "time_elapsed": ["20:28:52", "23 days, 0:30:43"],
            "T": [2, 5],
        },
        index=[0, 1],
    )
    assert info_df.equals(sweep_data.info())

    assert sweep_data.get(T=2).out == sweep_dir / "0-run"

    T = me.T([2, 5])
    Ms = me.Ms([6781.89022085, 6810.43736377])
    Cv = me.Entity("IsochoricHeatCapacity", [0, 0])
    E = me.Entity("Energy", [0, 0])
    U_binder = np.array([0.50682319, 0.50969701])
    assert T == sweep_data.T
    assert Ms == sweep_data.Ms
    assert Cv == sweep_data.Cv
    assert E == sweep_data.E
    assert np.allclose(U_binder, sweep_data.U_binder)


def test_TemperatureSweepData_output(DATA, tmp_path):
    sweep_data = uppasd.read(DATA / "uppasd" / "0-temperature_sweep")
    sweep_data.save_output(tmp_path)
    collection = me.io.entities_from_file(tmp_path / "output.csv")
    T = me.T([2, 5])
    Ms = me.Ms([6781.89022085, 6810.43736377])
    Cv = me.Entity("IsochoricHeatCapacity", [0, 0])
    U_binder = np.array([0.50682319, 0.50969701])
    E = me.Entity("Energy", [0, 0])
    assert collection.T == T
    assert collection.Ms == Ms
    assert collection.Cv == Cv
    assert np.allclose(collection.U_binder, U_binder)
    assert collection.E == E


def test_TemperatureSweepData_yaml(DATA):
    """Test TemperatureSweepData yaml file."""
    with open(
        DATA / "uppasd" / "0-temperature_sweep" / "mammos_spindynamics.yaml"
    ) as f:
        info = yaml.safe_load(f)
    assert info == {
        "metadata": {
            "index": 0,
            "description": "Test sweep.",
            "UppASD version": "v6.0.2",
            "mammos_spindynamics_version": "0.2.5",
            "mode": "temperature_sweep",
        },
        "parameters": {"T": [2, 5]},
    }



"""UppASD Data class."""

from __future__ import annotations

import pathlib
import re
from io import StringIO
from typing import TYPE_CHECKING

import mammos_entity as me
import mammos_units as u
import numpy as np
import pandas as pd
import yaml

if TYPE_CHECKING:
    import mammos_entity
    import numpy
    import pandas


def read(out: pathlib.Path | str) -> MammosUppasdData | RunData | TemperatureSweepData:
    """Read UppASD calculations results directory."""
    out = pathlib.Path(out)
    if not out.is_dir():
        raise RuntimeError(f"Output directory {out} does not exist.")

    if (info_yaml := out / "mammos_spindynamics.yaml").is_file():
        with open(info_yaml) as f:
            info = yaml.safe_load(f)
        if info["metadata"]["mode"] == "mammos_uppasd_data":
            return MammosUppasdData(out)
        elif info["metadata"]["mode"] == "run":
            return RunData(out)
        elif info["metadata"]["mode"] == "temperature_sweep":
            return TemperatureSweepData(out)
        else:
            mode = info["metadata"]["mode"]
            raise RuntimeError(
                f"Unable to understand mode: '{mode}' in mammos_spindynamics.yaml. "
            )
    else:
        raise RuntimeError(f"mammos_spindynamics.yaml file not found in path: {out}.")


class MammosUppasdData:
    """Collection of UppASD Result instances."""

    def __init__(self, out: pathlib.Path | str):
        """Initialize MammosUppasdData given the directory containing all runs."""
        self.out = pathlib.Path(out)

    def __repr__(self):
        """Define repr."""
        return f"MammosUppasdData('{self.out}')"

    def __getitem__(self, idx):
        """Extract i-th run."""
        runs = [
            run_dir
            for run_dir in self.out.iterdir()
            if re.match(r"\d+-(run|temperature_sweep)$", run_dir.name)
        ]
        runs.sort(key=lambda x: int(x.name.split("-")[0]))
        return read(runs[idx])

    def get(self, **kwargs):
        """Extract run based on different filters."""
        df = self.info()
        for key, value in kwargs.items():
            df = df[df[key] == value]
        if len(df) == 0:
            raise LookupError(
                f"Requested run not found.\nSelection: {kwargs}\n"
                f"Available data: {self.info()}"
            )
        if len(df) > 1:
            error_string = (
                "Too many results. Please refine your search.\n"
                + "Avilable materials based on request:\n"
            )
            error_string += str(df)
            raise LookupError(error_string)
        return read(self.out / df.iloc[0]["name"])

    def info(
        self,
        include_description: bool = True,
        include_time_elapsed: bool = True,
        include_parameters: bool = True,
    ) -> pandas.DataFrame:
        """Dataframe containing information about all available UppASD runs."""
        metadata_keys = []
        if include_description:
            metadata_keys += ["description"]
        if include_time_elapsed:
            metadata_keys += ["time_elapsed"]
        all_runs = []
        index = []
        for run in self:
            if run:
                info_ = {"name": run.out.name}
                index.append(int(run.out.name.split("-")[0]))
                if metadata_keys:
                    metadata_ = {k: run.metadata.get(k) for k in metadata_keys}
                    info_ = {**info_, **metadata_}
                if include_parameters:
                    info_ = {**info_, **run.parameters}
                all_runs.append(info_)
        return pd.DataFrame(all_runs, index=index)


class RunData:
    """UppASD Data parser class for a single run."""

    def __init__(self, out: pathlib.Path):
        """Initialize Data given the output directory of a single run."""
        self.out = pathlib.Path(out)
        with open(self.out / "mammos_spindynamics.yaml") as f:
            info = yaml.safe_load(f)
        self._input_dictionary = _parse_inpsd_file(self.inpsd)
        self.metadata = info["metadata"]
        self.parameters = info["parameters"]
        df = pd.read_csv(self.cumulants, sep=r"\s+")
        self._cumulant_data = df.iloc[-1]

    def __repr__(self):
        """Define repr."""
        return f"RunData('{self.out}')"

    def info(
        self,
        include_description: bool = True,
        include_time_elapsed: bool = True,
        include_parameters: bool = True,
    ) -> pandas.DataFrame:
        """Return information about the UppASD run."""
        out = {"name": [self.out.name]}
        if include_description:
            out["description"] = [self.metadata["description"]]
        if include_time_elapsed:
            out["time_elapsed"] = [self.metadata["time_elapsed"]]
        if include_parameters:
            parameters_dictionary = {k: [v] for k, v in self.parameters.items()}
            out = {**out, **parameters_dictionary}
        return pd.DataFrame(out)

    @property
    def T(self) -> float:
        """Get Thermodynamics Temperature."""
        return me.T(self._input_dictionary["temp"])

    @property
    def inpsd(self) -> pathlib.Path:
        """Get path of ``inpsd.dat`` input file."""
        return pathlib.Path(self.out / "inpsd.dat")

    @property
    def exchange(self) -> pathlib.Path:
        """Get path of file containing exchange interactions."""
        return self.out / self._input_dictionary["exchange"]

    @property
    def momfile(self) -> pathlib.Path:
        """Get path of file containing magnetic moments."""
        return self.out / self._input_dictionary["momfile"]

    @property
    def posfile(self) -> pathlib.Path:
        """Get path of file containing atomic positions."""
        return self.out / self._input_dictionary["posfile"]

    @property
    def cumulants(self) -> pathlib.Path:
        """Get ``cumulants.*.out`` file."""
        cumulants = list(self.out.glob("cumulants.*.out"))
        if not cumulants:
            raise RuntimeError(f"Could not find a cumulants.<simid>.out in {self.out}")
        if len(cumulants) > 1:
            # we assume that this will never happen
            raise RuntimeError(
                f"Found multiple candidates for cumulants.<simid>.out in {self.out}:\n"
                f"{cumulants}"
            )
        return cumulants[0]

    @property
    def restartfile(self) -> pathlib.Path:
        """Get path of restart file."""
        restart_files = list(self.out.glob("restart.*.out"))
        if not restart_files:
            raise RuntimeError(f"Could not find restart.<simid>.out in {self.out}")
        if len(restart_files) > 1:
            # we assume that this will never happen
            raise RuntimeError(
                f"Found multiple candidates for restart.<simid>.out in {self.out}:\n"
                f"{restart_files}"
            )
        return restart_files[0]

    @property
    def n_magnetic_atoms(self) -> int:
        """Get number of magnetic atoms."""
        with open(self.momfile, encoding="utf-8") as file:
            n = len(file.readlines())
        return n

    @property
    def Ms(self) -> mammos_entity.Entity:
        """Get Spontaneous Magnetization."""
        cell = self._input_dictionary["cell"]
        lattice_const = self._input_dictionary["alat"] * u.m
        cell_volume = np.dot(cell[0], np.cross(cell[1], cell[2])) * lattice_const**3
        Ms_mu_B_per_atom = float(self._cumulant_data["<M>"]) * u.mu_B
        Ms = Ms_mu_B_per_atom * self.n_magnetic_atoms / cell_volume
        return me.Ms(Ms, unit="A/m")

    @property
    def Cv(self) -> mammos_entity.Entity:
        """Get specific heat capacity."""
        k_B = u.constants.k_B
        Cv = float(self._cumulant_data["C_v(tot)"]) * k_B * self.n_magnetic_atoms
        return me.Entity("IsochoricHeatCapacity", Cv)

    @property
    def E(self) -> mammos_entity.Entity:
        """Get energy."""
        E = float(self._cumulant_data["<E>"]) * u.mRy * self.n_magnetic_atoms
        return me.Entity("Energy", E, unit="J")

    @property
    def U_binder(self) -> float:
        """Get U_binder value."""
        U_b = float(self._cumulant_data["U_{Binder}"])
        return U_b


class TemperatureSweepData:
    """Class for the result of the temperature_array runner."""

    def __init__(self, out: pathlib.Path | str):
        """Initialize TemperatureSweepData given the output directory of a sweep run."""
        self.out = pathlib.Path(out)
        with open(self.out / "mammos_spindynamics.yaml") as f:
            info = yaml.safe_load(f)
        self.metadata = info["metadata"]
        self.parameters = info["parameters"]

    def __repr__(self):
        """Define repr."""
        return f"TemperatureSweepData('{self.out}')"

    def __getitem__(self, idx):
        """Extract i-th run."""
        runs = [
            run_dir
            for run_dir in self.out.iterdir()
            if re.match(r"\d+-run", run_dir.name)
        ]
        return read(runs[idx])

    def info(
        self,
        include_description: bool = True,
        include_time_elapsed: bool = True,
        include_parameters: bool = True,
    ) -> pandas.DataFrame:
        """Dataframe containing information about all available UppASD runs."""
        metadata_keys = []
        if include_description:
            metadata_keys += ["description"]
        if include_time_elapsed:
            metadata_keys += ["time_elapsed"]
        all_runs = []
        index = []
        for run in self:
            if run:
                info_ = {"name": run.out.name}
                index.append(
                    int(
                        run.out.name.replace("-run", "").replace(
                            "-temperature_sweep", ""
                        )
                    )
                )
                if metadata_keys:
                    metadata_ = {k: run.metadata[k] for k in metadata_keys}
                    info_ = {**info_, **metadata_}
                if include_parameters:
                    info_ = {**info_, **run.parameters}
                all_runs.append(info_)
        return pd.DataFrame(all_runs, index=index)

    def get(self, **kwargs) -> RunData:
        """Select run satisfying certain filters defined in the keyword arguments.

        If there are multiple matches, it returns the first one.
        """
        df = self.info()
        for key, val in kwargs.items():
            df = df[df[key] == val]
        if len(df) == 0:
            raise LookupError(
                f"Requested run not found.\nSelection: {kwargs}\n"
                f"Available data: {self.info()}"
            )
        if len(df) > 1:
            error_string = (
                "Too many results. Please refine your search.\n"
                + "Avilable materials based on request:\n"
            )
            error_string += str(df)
            raise LookupError(error_string)
        return read(self.out / df.iloc[0]["name"])

    @property
    def T(self) -> mammos_entity.Entity:
        """Get Thermodynamics Temperature."""
        return me.concat_flat(*[run.T for run in self if run])

    @property
    def Ms(self) -> mammos_entity.Entity:
        """Get Spontaneous Magnetization."""
        return me.concat_flat(*[run.Ms for run in self if run])

    @property
    def Cv(self) -> mammos_entity.Entity:
        """Get Isochoric Heat Capacity.

        The isochorich (at constant volume) heat capacity Cv is evaluated as the
        derivative of the energy as a function of temperature.
        """
        k_B = u.constants.k_B.value
        Cv = np.gradient(self.E.value / k_B, self.T.value, axis=0)
        return me.Entity("IsochoricHeatCapacity", Cv)

    @property
    def U_binder(self) -> numpy.ndarray:
        """Get Binder coefficient."""
        return np.array([run.U_binder for run in self if run])

    @property
    def E(self) -> mammos_entity.Entity:
        """Get Energy."""
        return me.concat_flat(*[run.E for run in self if run])

    def save_output(self, out: pathlib.Path | str) -> None:
        """Save output files M(T) and output.csv in directory `out`.

        `M(T)` contains all information evaluated from the cumulant files.
        `output.csv` contains entities for information temperature,
        magnetization, Binder cumulant and heat capacity.
        """
        out = pathlib.Path(out)
        out.mkdir(parents=True, exist_ok=True)

        with open(self[0].cumulants) as f:
            lines = f.readlines()
        header = lines[0]

        with open(out / "M(T)", "w") as f:
            f.write(f"{'T':>5} {header}")
            for run in self:
                if run:
                    with open(run.cumulants) as f_run:
                        lines = f_run.readlines()
                    f.write(f"{run.T.value:>5.0f} {lines[-1]}")

        me.io.entities_to_file(
            out / "output.csv",
            "Magnetization and heat capacity from UppASD",
            T=self.T,
            Ms=self.Ms,
            U_binder=self.U_binder,
            Cv=self.Cv,
            E=self.E,
        )


def _parse_inpsd_file(inpsd_file: pathlib.Path | str) -> dict:
    """Parse inpsd.dat file."""
    string_parameters = {
        "simid",
        "exchange",
        "momfile",
        "posfile",
        "restartfile",
    }
    float_parameters = {
        "alat",
        "temp",
    }
    with open(inpsd_file, encoding="utf-8") as file:
        lines = file.readlines()
    parameters = {}
    for i, line in enumerate(lines):
        for par in string_parameters:
            if re.match(f"{par} .*", line):
                parameters[par] = line.split()[1]
        for par in float_parameters:
            if re.match(f"{par} .*", line):
                try:
                    parameters[par] = float(line.split()[1])
                except ValueError:
                    continue
        if re.match("(bc|BC) .*", line):
            parameters["bc"] = line.removeprefix("bc").split()[:3]
        if re.match("cell .*", line):
            a = np.genfromtxt(StringIO(line.removeprefix("cell")))
            b = np.genfromtxt(StringIO(lines[i + 1]))
            c = np.genfromtxt(StringIO(lines[i + 2]))
            parameters["cell"] = np.vstack((a, b, c))
        if re.match("ncell .*", line):
            parameters["ncell"] = np.genfromtxt(StringIO(line.removeprefix("ncell")))[
                :3
            ]
    return parameters



"""Functions for reading tables."""

from __future__ import annotations

import pathlib
from textwrap import dedent
from typing import TYPE_CHECKING

import mammos_entity as me
import mammos_units as u
import matplotlib.pyplot as plt
import numpy as np
import pandas
import pandas as pd
from pydantic import ConfigDict
from pydantic.dataclasses import dataclass
from rich import print

if TYPE_CHECKING:
    import matplotlib

DATA_DIR = pathlib.Path(__file__).parent / "data"


def _check_short_label(short_label: str) -> str | int:
    """Check that short label follows the standards and returns material parameters.

    Args:
        short_label: Short label containing chemical formula and space group
    number separated by a hyphen.

    Returns:
        Chemical formula and space group number.

    Raises:
        ValueError: Wrong format.

    """
    short_label_list = short_label.split("-")
    if len(short_label_list) != 2:
        raise ValueError(
            dedent(
                """
                Wrong format for `short_label`.
                Please use the format <chemical_formula>-<space_group_number>.
                """
            )
        )
    chemical_formula = short_label_list[0]
    space_group_number = int(short_label_list[1])
    return chemical_formula, space_group_number


def get_spontaneous_magnetization(
    chemical_formula: str | None = None,
    space_group_name: str | None = None,
    space_group_number: int | None = None,
    cell_length_a: float | None = None,
    cell_length_b: float | None = None,
    cell_length_c: float | None = None,
    cell_angle_alpha: float | None = None,
    cell_angle_beta: float | None = None,
    cell_angle_gamma: float | None = None,
    cell_volume: float | None = None,
    ICSD_label: str | None = None,
    OQMD_label: str | None = None,
    jfile=None,
    momfile=None,
    posfile=None,
    print_info: bool = False,
) -> MagnetizationData:
    """Get spontaneous magnetization interpolator from a database.

    This function retrieves the temperature-dependent spontaneous magnetization
    from a local database of spin dynamics calculations.
    Data is retrieved by querying material information or UppASD input files.

    Args:
        chemical_formula: Chemical formula
        space_group_name: Space group name
        space_group_number: Space group number
        cell_length_a: Cell length x
        cell_length_b: Cell length y
        cell_length_c: Cell length z
        cell_angle_alpha: Cell angle alpha
        cell_angle_beta: Cell angle beta
        cell_angle_gamma: Cell angle gamma
        cell_volume: Cell volume
        ICSD_label: Label in the NIST Inorganic Crystal Structure Database.
        OQMD_label: Label in the the Open Quantum Materials Database.
        jfile: TODO
        momfile: TODO
        posfile: TODO
        print_info: Whether to print information about the retrieved material.

    Returns:
        Interpolator function based on available data.

    Examples:
        >>> import mammos_spindynamics.db
        >>> mammos_spindynamics.db.get_spontaneous_magnetization("Nd2Fe14B")
        MagnetizationData(T=..., Ms=...)

    """
    if posfile is not None:
        table = _load_uppasd_simulation(
            jfile=jfile, momfile=momfile, posfile=posfile, print_info=print_info
        )
    else:
        table = _load_ab_initio_data(
            print_info=print_info,
            chemical_formula=chemical_formula,
            space_group_name=space_group_name,
            space_group_number=space_group_number,
            cell_length_a=cell_length_a,
            cell_length_b=cell_length_b,
            cell_length_c=cell_length_c,
            cell_angle_alpha=cell_angle_alpha,
            cell_angle_beta=cell_angle_beta,
            cell_angle_gamma=cell_angle_gamma,
            cell_volume=cell_volume,
            ICSD_label=ICSD_label,
            OQMD_label=OQMD_label,
        )

    return MagnetizationData(
        me.Entity("ThermodynamicTemperature", value=table["T[K]"], unit=u.K),
        me.Ms(table["M[A/m]"], unit=u.A / u.m),
    )


@dataclass(config=ConfigDict(arbitrary_types_allowed=True, frozen=True))
class MagnetizationData:
    """Magnetization data.

    Contains temperature and spontaneous magnetization data.
    """

    T: me.Entity
    """Array of temperatures."""
    Ms: me.Entity
    """Array of spontaneous magnetizations for the different temperatures."""

    @property
    def dataframe(self):
        """Dataframe containing temperature and spontaneous magnetization data."""
        return pd.DataFrame(
            {
                "T": self.T.value,
                "Ms": self.Ms.value,
            }
        )

    def plot(
        self, ax: matplotlib.axes.Axes | None = None, **kwargs
    ) -> matplotlib.axes.Axes:
        """Plot the spontaneous magnetization data-points."""
        if not ax:
            _, ax = plt.subplots()
        kwargs.setdefault("marker", "x")
        ax.plot(self.T.value, self.Ms.value, linestyle="", **kwargs)
        ax.set_xlabel(self.T.axis_label)
        ax.set_ylabel(self.Ms.axis_label)
        if "label" in kwargs:
            ax.legend()
        return ax


def _load_uppasd_simulation(
    jfile: str | pathlib.Path,
    momfile: str | pathlib.Path,
    posfile: str | pathlib.Path,
    print_info: bool = False,
) -> pandas.DataFrame:
    """Find UppASD simulation results with given input files in database.

    Args:
        jfile: Location of `jfile`
        momfile: Location of `momfile`.
        posfile: Location of `posfile`.
        print_info: Whether to print information about the retrieved material.

    Returns:
        Table of pre-calculated Temperature-dependent Magnetization values.

    Raises:
        LookupError: Simulation not found in database.

    """
    j = _parse_jfile(jfile)
    mom = _parse_momfile(momfile)
    pos = _parse_posfile(posfile)
    for ii in DATA_DIR.iterdir():
        if ii.is_dir() and _check_input_files(ii, j, mom, pos):
            table = pd.read_csv(ii / "M.csv", header=[0, 1])
            if print_info:
                print("Found material in database.")
                print(_describe_material(material_label=ii.name))
            return table
    raise LookupError("Requested simulation not found in database.")


def _parse_jfile(jfile: str | pathlib.Path) -> pandas.DataFrame:
    """Parse jfile, input for UppASD.

    See https://uppasd.github.io/UppASD-manual/input/#exchange
    for the correct formatting.

    Args:
        jfile: Location of `jfile`.

    Returns:
        Dataframe of exchange interactions.

    Raises:
        SyntaxError: Wrong formatting.

    """
    with open(jfile) as ff:
        jlines = ff.readlines()
    try:
        df = pd.DataFrame(
            [
                [int(x) for x in li[:-1]] + [float(x) for x in li[-1:]]
                for li in [line.split() for line in jlines]
            ],
            columns=[
                "atom_i",
                "atom_j",
                "interaction_x",
                "interaction_y",
                "interaction_z",
                "exchange_energy[mRy]",
            ],
        ).sort_values(
            by=["atom_i", "atom_j", "interaction_x", "interaction_y", "interaction_z"],
        )
        return df
    except ValueError:
        raise SyntaxError(
            dedent(
                """
                Unable to parse jfile.
                Please check syntax according to
                https://uppasd.github.io/UppASD-manual/input/#exchange
                """
            )
        ) from None


def _parse_momfile(momfile: str | pathlib.Path) -> dict:
    """Parse momfile, input for UppASD.

    See https://uppasd.github.io/UppASD-manual/input/#momfile
    for the correct formatting.

    Args:
        momfile: Location of `momfile`.

    Returns:
        Dictionary of magnetic moment information.

    Raises:
        SyntaxError: Wrong formatting.

    """
    with open(momfile) as ff:
        momlines = ff.readlines()
    try:
        mom = {
            int(line[0]): {
                "chemical_type": int(line[1]),
                "magnetic_moment_magnitude[muB]": float(line[2]),
                "magnetic_moment_direction": np.array([float(x) for x in line[3:]]),
            }
            for line in [ll.split() for ll in momlines]
        }
        return mom
    except ValueError:
        raise SyntaxError(
            dedent(
                """
                Unable to parse momfile.
                Please check syntax according to
                https://uppasd.github.io/UppASD-manual/input/#momfile
                """
            )
        ) from None


def _parse_posfile(posfile: str | pathlib.Path) -> dict:
    """Parse posfile, input for UppASD.

    See https://uppasd.github.io/UppASD-manual/input/#posfile
    for the correct formatting.

    Args:
        posfile: Location of `posfile`.

    Returns:
        Dictionary of atoms position information.

    Raises:
        SyntaxError: Wrong formatting.

    """
    with open(posfile) as ff:
        poslines = ff.readlines()
    try:
        pos = {
            int(line[0]): {
                "atom_type": int(line[1]),
                "atom_position": np.array([float(x) for x in line[2:]]),
            }
            for line in [ll.split() for ll in poslines]
        }
        return pos
    except ValueError:
        raise SyntaxError(
            dedent(
                """
                Unable to parse posfile.
                Please check syntax according to
                https://uppasd.github.io/UppASD-manual/input/#posfile
                """
            )
        ) from None


def _check_input_files(
    dir_i: pathlib.Path, j: pandas.DataFrame, mom: dict, pos: dict
) -> bool:
    """Check if UppASD inputs are equivalent to the ones in directory `dir_i`.

    The extracted input information `j`, `mom`, and `pos` are compared with the
    extracted information from the files in directory `dir_i`.
    If the inputs are close enough, this function returns `True`.

    Args:
        dir_i: Considered directory in the database
        j: Dataframe of exchange interactions.
        mom: Dictionary of magnetic moment information.
        pos: Dictionary of atoms position information.

    Returns:
        bool: True` if the inputs match almost exactly. `False` otherwise.

    """
    if not (
        (dir_i / "jfile").is_file()
        or (dir_i / "momfile").is_file()
        or (dir_i / "posfile").is_file()
    ):
        return False

    if (dir_i / "jfile").is_file():
        j_i = _parse_jfile(dir_i / "jfile")
        if not j_i.drop("exchange_energy[mRy]", axis=1).equals(
            j.drop("exchange_energy[mRy]", axis=1)
        ):
            return False
        if not np.allclose(
            j_i["exchange_energy[mRy]"].to_numpy(),
            j["exchange_energy[mRy]"].to_numpy(),
        ):
            return False

    if (dir_i / "momfile").is_file():
        mom_i = _parse_momfile(dir_i / "momfile")
        if len(mom_i) != len(mom):
            return False
        for index, site in mom_i.items():
            if (
                site["chemical_type"] != mom[index]["chemical_type"]
                or not np.allclose(
                    site["magnetic_moment_magnitude[muB]"],
                    mom[index]["magnetic_moment_magnitude[muB]"],
                )
                or not np.allclose(
                    site["magnetic_moment_direction"],
                    mom[index]["magnetic_moment_direction"],
                )
            ):
                return False

    if (dir_i / "posfile").is_file():
        pos_i = _parse_posfile(dir_i / "posfile")
        if len(pos_i) != len(pos):
            return False
        for index, atom in pos_i.items():
            if atom["atom_type"] != pos[index]["atom_type"] or not np.allclose(
                atom["atom_position"],
                pos[index]["atom_position"],
            ):
                return False

    return True


def _load_ab_initio_data(print_info: bool = False, **kwargs) -> pandas.DataFrame:
    """Load material with given structure information.

    Args:
        print_info: Print info
        **kwargs: Selection arguments

    Returns:
        Table of pre-calculated Temperature-dependent Magnetization values.

    Raises:
        LookupError: Requested material not found in database.
        LookupError: Too many results found with this formula.

    """
    df = find_materials(**kwargs)
    num_results = len(df)
    if num_results == 0:
        raise LookupError("Requested material not found in database.")
    elif num_results > 1:  # list all possible choice
        error_string = (
            "Too many results. Please refine your search.\n"
            + "Avilable materials based on request:\n"
        )
        for _row, material in df.iterrows():
            error_string += _describe_material(material)
        raise LookupError(error_string)

    material = df.iloc[0]
    if print_info:
        print("Found material in database.")
        print(_describe_material(material))
    return pd.read_csv(DATA_DIR / material.label / "M.csv")


def find_materials(**kwargs) -> pandas.DataFrame:
    """Find materials in database.

    This function retrieves one or known materials from the database
    `db.csv` by filtering for any requirements given in `kwargs`.

    Args:
        kwargs: Selection arguments

    Returns:
        Dataframe containing materials with requested qualities. Possibly empty.

    """
    df = pd.read_csv(
        DATA_DIR / "db.csv",
        converters={
            "chemical_formula": str,
            "space_group_name": str,
            "space_group_number": int,
            "cell_length_a": u.Quantity,
            "cell_length_b": u.Quantity,
            "cell_length_c": u.Quantity,
            "cell_angle_alpha": u.Quantity,
            "cell_angle_beta": u.Quantity,
            "cell_angle_gamma": u.Quantity,
            "cell_volume": u.Quantity,
            "ICSD_label": str,
            "OQMD_label": str,
            "label": str,
        },
        comment="#",
    )
    for key, value in kwargs.items():
        if value is not None:
            if isinstance(value, u.Quantity):
                df = df[df[key] == value.to(df[key].unit)]
            else:
                df = df[df[key] == value]
    return df


def _describe_material(
    material: pandas.DataFrame | None = None, material_label: str | None = None
) -> str:
    """Describe material in a complete way.

    This function returns a string listing the properties of the given material
    or the given material label.

    Args:
        material: Material dataframe containing structure information.
        material_label: Label of material in local database.

    Returns:
        Description of the material.

    Raises:
        ValueError: If material and material label are both None.

    """
    if material is None and material_label is None:
        raise ValueError("Material and material label cannot be both empty.")
    if material_label is not None:
        df = find_materials()
        material = df[df["label"] == material_label].iloc[0]
    return dedent(
        f"""
            Chemical Formula: {material.chemical_formula}
            Space group name: {material.space_group_name}
            Space group number: {material.space_group_number}
            Cell length a: {material.cell_length_a}
            Cell length b: {material.cell_length_b}
            Cell length c: {material.cell_length_c}
            Cell angle alpha: {material.cell_angle_alpha}
            Cell angle beta: {material.cell_angle_beta}
            Cell angle gamma: {material.cell_angle_gamma}
            Cell volume: {material.cell_volume}
            ICSD_label: {material.ICSD_label}
            OQMD_label: {material.OQMD_label}
            """
    )
